#include <atlbase.h>
#include <sven_ui/commands.h>
#include <sven_ui/template_command.h>

#include <sven/gl_array.h>
#include <nehe/matrix_stack.h>

#include <qedit.h>

#include "dx_cap.h"
#include <stdexcept>
#include <sven/fmt.h>
#include <sven/string_itr.h>
#include <sven/foreach.h>

#include "process_frame.h"

using sven::qw;

#pragma comment( lib, "quartz.lib " )
#pragma comment( lib, "odbc32.lib" )
#pragma comment( lib, "odbccp32.lib" )

//stupid direct show base classes... used in all the demo's but not *actually* part of the library.
//what rare genius thought that would be a good idea, i wonder...
#ifdef _DEBUG
#pragma comment( lib, "strmbasd.lib" )
#else
#pragma comment( lib, "strmbase.lib" )
#endif

//#pragma comment( lib, "C:\\Program Files\\Microsoft DirectX 9.0 SDK (Summer 2004)\\Samples\\C++\\DirectShow\\BaseClasses\\Debug\\strmbasd.lib" )


//kinda pondering these macro's, i mean, i could probably do something like this by extending format somehow,
//but it would be hard, and kinda hacky, and really, why bother?  macro's are just a cleaner solution in this case.

#define RT_ERROR_ON_FAILED(x) if(FAILED(x)) \
				throw std::runtime_error((fmt( \
				"The following microsoft SDK command, %s, has failed, "\
				"which your programmer believes is worthy of a runtime error") \
				% qw(#x)).str());

//
// Macros
//
#define SAFE_RELEASE(x) { if (x) x->Release(); x = NULL; }

#define JIF(x) if (FAILED(hr=(x))) \
    {throw std::runtime_error((fmt("FAILED(hr=0x%x) in " TEXT(#x) TEXT("\n\0") % hr).str()); return hr;}

//
// Constants
//
#define DEFAULT_VIDEO_WIDTH     320
#define DEFAULT_VIDEO_HEIGHT    320

#define APPLICATIONNAME TEXT("Video Capture Previewer (PlayCap)\0")
#define CLASSNAME       TEXT("VidCapPreviewer\0")

// Application-defined message to notify app of filtergraph events
#define WM_GRAPHNOTIFY  WM_APP+1

//
// Resource constants
//
#define IDI_VIDPREVIEW          100

namespace direct_show {

using sven::cstring;

ISampleGrabber * sample_grabber = NULL;
//bool disable_grabbing=true;

struct GL_GrabberCB:  public ISampleGrabberCB {
	long frame_count;

	GL_GrabberCB() : frame_count(0) {}
	//~GL_GrabberCB() {}
	
	//right now, just using the buffer thingy...
    STDMETHODIMP SampleCB(double SampleTime, IMediaSample *pSample) 
    {return E_NOTIMPL;
    
    //sven_ui::report("yo.");
    //return S_OK;
    }
    
    STDMETHODIMP BufferCB(double SampleTime, BYTE *pBuffer, long BufferLen) {
	//	if(disable_grabbing) return S_OK;
		try {
		//sven_ui::report("yo.");
		frame_count++;




		

		//blit the data.
	//	const int width=200;
	//	const int height=200;
//		glRasterPos2f(0,0);
//		glDrawPixels(width,height,GL_RGB,GL_BYTE,pBuffer);
	
//		glColor3f(0,1,0);
//		glBox(sven::rect(0,0,width,height));

		process_frame::send_data(pBuffer);
		
		
		} catch (std::exception &e) {
		MessageBox(NULL,e.what(),"Exception caught in BufferCB",MB_OK|MB_ICONEXCLAMATION);
		}
	
		
		return S_OK ;
    }
		
	STDMETHODIMP QueryInterface(REFIID riid, void ** ppv) {
        if( riid == IID_ISampleGrabberCB || riid == IID_IUnknown ) {
            *ppv = (void *) static_cast<ISampleGrabberCB*> ( this );
            return NOERROR;
        }    
        return E_NOINTERFACE;
    }

    // fake out any COM ref counting, 
    // we'll mannage this memory ourselves.
    STDMETHODIMP_(ULONG) AddRef() { return 2; }
    STDMETHODIMP_(ULONG) Release() { return 1; }	

};



inline void Msg(cstring s) {
	throw std::runtime_error(s);
}
inline void Msg(cfmt f) { Msg(f.str()); } 

// An application can advertise the existence of its filter graph
// by registering the graph with a global Running Object Table (ROT).
// The GraphEdit application can detect and remotely view the running
// filter graph, allowing you to 'spy' on the graph with GraphEdit.
//
// To enable registration in this sample, define REGISTER_FILTERGRAPH.
//
//#define REGISTER_FILTERGRAPH


//
// Global data
//
HWND ghApp=0;
DWORD g_dwGraphRegister=0;

IMediaControl * g_pMC = NULL;
IMediaEventEx * g_pME = NULL;
IGraphBuilder * g_pGraph = NULL;
ICaptureGraphBuilder2 * g_pCapture = NULL;
PLAYSTATE g_psCurrent = Stopped;

//	m_pSampleGrabber->SetCallback(m_pTrackingCB, 0);
// Set the media type (needed to get the bmp header info
//	m_pSampleGrabber->GetConnectedMediaType(&mt);
//	m_pTrackingCB->SetGrabMediaType(mt);


IBaseFilter *null_render_filter=NULL;
IBaseFilter *sample_grabber_filter=NULL;

GL_GrabberCB *	grabber_cb = NULL;

HRESULT CaptureVideo()
{
    HRESULT hr;
    IBaseFilter *pSrcFilter=NULL;

    // Get DirectShow interfaces
    hr = GetInterfaces();
    if (FAILED(hr))
    {
        Msg(fmt("Failed to get video interfaces!  hr=0x%x") % hr);
        return hr;
    }

	sven_ui::report("got the interfaces.");

    // Attach the filter graph to the capture graph
    hr = g_pCapture->SetFiltergraph(g_pGraph);
    if (FAILED(hr))
    {
        Msg(fmt("Failed to set capture filter graph!  hr=0x%x") % hr);
        return hr;
    }

    // Use the system device enumerator and class enumerator to find
    // a video capture/preview device, such as a desktop USB video camera.
    hr = FindCaptureDevice(&pSrcFilter);
    if (FAILED(hr))
    {
		RT_ERROR("stuff broke");
        // Don't display a message because FindCaptureDevice will handle it
        return hr;
    }
   
    // Add Capture filter to our graph.
    hr = g_pGraph->AddFilter(pSrcFilter, L"Video Capture");
    if (FAILED(hr))
    {
        Msg(fmt("Couldn't add the capture filter to the graph!  hr=0x%x\r\n\r\n" 
				"If you have a working video capture device, please make sure\r\n"
				"that it is connected and is not being used by another application.\r\n\r\n"
				"The sample will now close.") % hr);
        pSrcFilter->Release();
        return hr;
    }


	//create the filter object and put it in the graph
	{
		//create the filter
		RT_ERROR_ON_FAILED(
			CoCreateInstance(CLSID_SampleGrabber, NULL, CLSCTX_INPROC_SERVER,
			IID_IBaseFilter, reinterpret_cast<void**>(&sample_grabber_filter))
		);
	
		RT_ERROR_ON_FAILED(
			sample_grabber_filter->QueryInterface(IID_ISampleGrabber,
			    reinterpret_cast<void**>(&sample_grabber))
		);
	
	
		RT_ERROR_ON_FAILED(
			g_pGraph->AddFilter(sample_grabber_filter, L"SampleGrabber")
		);

	
	
		//setup the media type.
		//we happen to know that it's going to be 24bit video data,
		//which saves us the pain and suffering involved in trying to get
		//that info by hand.
		AM_MEDIA_TYPE mt;
		ZeroMemory(&mt, sizeof(AM_MEDIA_TYPE));
		mt.majortype	= MEDIATYPE_Video;
		mt.subtype		= MEDIASUBTYPE_RGB24;

		RT_ERROR_ON_FAILED(
			sample_grabber_filter->QueryInterface(IID_ISampleGrabber,
			reinterpret_cast<void**>(&sample_grabber))
		);

		RT_ERROR_ON_FAILED(
			sample_grabber->SetMediaType(&mt)
		);

		RT_ERROR_ON_FAILED(
			sample_grabber->SetBufferSamples(TRUE)
		);

		 


		//create our grabber callback object, and connect it to the sample_grabber.
		grabber_cb = new GL_GrabberCB();
		
		//turn on the buffer callback
		RT_ERROR_ON_FAILED(sample_grabber->SetCallback(grabber_cb, 1));	

	}

	// create and add the null render filter:
	// by using this when we render the stream, instead of the
	// default output filter, we will discard the data stream
	// instead of actually rendering it.
		
	hr = CoCreateInstance(CLSID_NullRenderer, NULL, CLSCTX_INPROC_SERVER,
		IID_IBaseFilter, (LPVOID *) &null_render_filter);
	if (FAILED(hr)) ("failed to create the null render filter");

	hr = g_pGraph->AddFilter(null_render_filter, L"NullRenderer");
	if (FAILED(hr)) RT_ERROR("failed to add the null render filter");

	

    // Render the preview pin on the video capture filter
    // Use this instead of g_pGraph->RenderFile
    hr = g_pCapture->RenderStream (
    &PIN_CATEGORY_CAPTURE
    /*&PIN_CATEGORY_PREVIEW*/, 
    &MEDIATYPE_Video,
    pSrcFilter, sample_grabber_filter, null_render_filter);
    
    if (FAILED(hr))
    {
        Msg(fmt("Couldn't render the video capture stream.  hr=0x%x\r\n"
				"The capture device may already be in use by another application.\r\n\r\n"
				"The sample will now close.") % hr);
        pSrcFilter->Release();
        return hr;
    }



    // Now that the filter has been added to the graph and we have
    // rendered its stream, we can release this reference to the filter.
    pSrcFilter->Release();
    


#ifdef REGISTER_FILTERGRAPH
    // Add our graph to the running object table, which will allow
    // the GraphEdit application to "spy" on our graph
    hr = AddGraphToRot(g_pGraph, &g_dwGraphRegister);
    if (FAILED(hr))
    {
        Msg(fmt("Failed to register filter graph with ROT!  hr=0x%x" % hr);
        g_dwGraphRegister = 0;
    }
#endif

//	hr = pF->QueryInterface(IID_ISampleGrabber,
  //  reinterpret_cast<void**>(&pGrabber));

    // Start previewing video data
    hr = g_pMC->Run();
    if (FAILED(hr))
    {
        Msg(fmt("Couldn't run the graph!  hr=0x%x") % hr);
        return hr;
    }


	static long garbage;
	//pMediaFilter->SetSyncSource(NULL); // Turn off the reference clock.
	//pControl->Run(); // Run the graph.
	//g_pME->WaitForCompletion(INFINITE, &garbage); 

    // Remember current state
    g_psCurrent = Running;
        
    
    {
		//send the dims of the stream being captured to the process frame functions
		
			AM_MEDIA_TYPE mt;
			RT_ERROR_ON_FAILED(	sample_grabber->GetConnectedMediaType(&mt));
			VIDEOINFOHEADER *pVih;
			if ((mt.formattype == FORMAT_VideoInfo) && 
				(mt.cbFormat >= sizeof(VIDEOINFOHEADER)) ) 
			{
				pVih = reinterpret_cast<VIDEOINFOHEADER*>(mt.pbFormat);
			}
			else RT_ERROR("something broke");
		
			int width=pVih->bmiHeader.biWidth; 
			int height=pVih->bmiHeader.biHeight; 	
    		process_frame::set_dims(width,height);
    }   
    return S_OK;
}



std::list<std::string> enumFilters(IGraphBuilder *pGraph) {

      IEnumFilters *pEnum = NULL;
      IBaseFilter *pFilter;
      ULONG cFetched;
      std::list<std::string> names;

      pGraph->EnumFilters(&pEnum);
      while(pEnum->Next(1, &pFilter, &cFetched) == S_OK)
      {
        FILTER_INFO FilterInfo;
        char szName[256];
        //std::string fname;

        pFilter->QueryFilterInfo(&FilterInfo);
        WideCharToMultiByte(CP_ACP, 0, FilterInfo.achName, -1, szName, 256, 0, 0);
        //fname= szName;
        names.push_back(szName);

		sven_ui::report(szName);

        SAFE_RELEASE(FilterInfo.pGraph);
        SAFE_RELEASE(pFilter);
      }

      SAFE_RELEASE(pEnum);

      return names;
}

//std::list<std::string> 

void reportFilters() {
	std::list<std::string> names = enumFilters(g_pGraph);
	std::list<std::string>::iterator i;
	//foreach(i,names) sven_ui::report(*i);
}

HRESULT FindCaptureDevice(IBaseFilter ** ppSrcFilter)
{
    HRESULT hr;
    IBaseFilter * pSrc = NULL;
    CComPtr <IMoniker> pMoniker =NULL;
    ULONG cFetched;

    if (!ppSrcFilter)
        return E_POINTER;
   
    // Create the system device enumerator
    CComPtr <ICreateDevEnum> pDevEnum =NULL;

    hr = CoCreateInstance (CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC,
                           IID_ICreateDevEnum, (void **) &pDevEnum);
    if (FAILED(hr))
    {
        Msg(fmt("Couldn't create system enumerator!  hr=0x%x") % hr);
        return hr;
    }

    // Create an enumerator for the video capture devices
    CComPtr <IEnumMoniker> pClassEnum = NULL;

    hr = pDevEnum->CreateClassEnumerator (CLSID_VideoInputDeviceCategory, &pClassEnum, 0);
    if (FAILED(hr))
    {
        Msg(fmt("Couldn't create class enumerator!  hr=0x%x") % hr);
        return hr;
    }

    // If there are no enumerators for the requested type, then 
    // CreateClassEnumerator will succeed, but pClassEnum will be NULL.
    if (pClassEnum == NULL)
    {
        RT_ERROR(  "No video capture device was detected.\n\n"
                   "This action requires a video capture device, such as a USB WebCam,"
                   "to be installed and working properly.");
        return E_FAIL;
    }

    // Use the first video capture device on the device list.
    // Note that if the Next() call succeeds but there are no monikers,
    // it will return S_FALSE (which is not a failure).  Therefore, we
    // check that the return code is S_OK instead of using SUCCEEDED() macro.
    if (S_OK == (pClassEnum->Next (1, &pMoniker, &cFetched)))
    {
        // Bind Moniker to a filter object
        hr = pMoniker->BindToObject(0,0,IID_IBaseFilter, (void**)&pSrc);
        if (FAILED(hr))
        {
            Msg(fmt("Couldn't bind moniker to filter object!  hr=0x%x") % hr);
            return hr;
        }
    }
    else
    {
        Msg(fmt("Unable to access video capture device!"));   
        return E_FAIL;
    }

    // Copy the found filter pointer to the output parameter.
    // Do NOT Release() the reference, since it will still be used
    // by the calling function.
    *ppSrcFilter = pSrc;

    return hr;
}


HRESULT GetInterfaces(void)
{
    HRESULT hr;

    // Create the filter graph
    hr = CoCreateInstance (CLSID_FilterGraph, NULL, CLSCTX_INPROC,
                           IID_IGraphBuilder, (void **) &g_pGraph);
    if (FAILED(hr))
        return hr;

    // Create the capture graph builder
    hr = CoCreateInstance (CLSID_CaptureGraphBuilder2 , NULL, CLSCTX_INPROC,
                           IID_ICaptureGraphBuilder2, (void **) &g_pCapture);
    if (FAILED(hr))
        return hr;
    
    // Obtain interfaces for media control and Video Window
    hr = g_pGraph->QueryInterface(IID_IMediaControl,(LPVOID *) &g_pMC);
    if (FAILED(hr))
        return hr;

//    hr = g_pGraph->QueryInterface(IID_IVideoWindow, (LPVOID *) &g_pVW);
//    if (FAILED(hr))
//       return hr;

    hr = g_pGraph->QueryInterface(IID_IMediaEvent, (LPVOID *) &g_pME);
    if (FAILED(hr))
        return hr;

    // Set the window handle used to process graph events
    hr = g_pME->SetNotifyWindow((OAHWND)ghApp, WM_GRAPHNOTIFY, 0);

    return hr;
}


void CloseInterfaces(void)
{
    // Stop previewing data
    if (g_pMC)
        g_pMC->StopWhenReady();

    g_psCurrent = Stopped;

    // Stop receiving events
    if (g_pME)
        g_pME->SetNotifyWindow(NULL, WM_GRAPHNOTIFY, 0);

    // Relinquish ownership (IMPORTANT!) of the video window.
    // Failing to call put_Owner can lead to assert failures within
    // the video renderer, as it still assumes that it has a valid
    // parent window.
//    if(g_pVW)
//    {
//        g_pVW->put_Visible(OAFALSE);
//        g_pVW->put_Owner(NULL);
//    }

#ifdef REGISTER_FILTERGRAPH
    // Remove filter graph from the running object table   
    if (g_dwGraphRegister)
        RemoveGraphFromRot(g_dwGraphRegister);
#endif

    // Release DirectShow interfaces
    SAFE_RELEASE(g_pMC);
    SAFE_RELEASE(g_pME);
//    SAFE_RELEASE(g_pVW);
    SAFE_RELEASE(g_pGraph);
    SAFE_RELEASE(g_pCapture);
}

bool ignore_data=true;

struct AcceptData : sven_ui::TemplateCommand {
	void fun(std::vector<Argument> &args, int n) {
		g_pMC->Run();
		/*disable_grabbing=false;
		RT_ERROR_ON_FAILED(
			sample_grabber->SetBufferSamples(TRUE)
		);	*/
		sven_ui::report("bitmap grabbing enabled");
	}

	AcceptData() : TemplateCommand("accept_data", "accept incomming data.", "no args", "general") {}
};

struct IgnoreData : sven_ui::TemplateCommand {
	void fun(std::vector<Argument> &args, int n) {

		g_pMC->Stop();
/*
		disable_grabbing=true;

		RT_ERROR_ON_FAILED(
			sample_grabber->SetBufferSamples(FALSE)
		);*/
		sven_ui::report("bitmap grabbing disabled");
	}

	IgnoreData() : TemplateCommand("ignore_data", "ignore incomming data.", "no args", "general") {}
};


void register_commands() {
	sven_ui::get_command_manager()->insert(new IgnoreData);
	sven_ui::get_command_manager()->insert(new AcceptData);	
}

}